# stdlib
import datetime
import logging
import time

# pyramid_debugtoolbar
from pyramid_debugtoolbar.panels import DebugPanel
from pyramid_debugtoolbar.utils import STATIC_PATH
from pyramid_debugtoolbar.utils import ROOT_ROUTE_NAME


# ==============================================================================


class DogpileDebugPanel(DebugPanel):
    """
    """
    name = 'Dogpile'
    template = 'pyramid_debugtoolbar_dogpile.panels:templates/pyramid_dogpile.dbtmako'
    title = 'Dogpile'
    nav_title = 'Dogpile'

    def __init__(self, request):
        self.logs = request.dogpile_logging['api_calls']

    @property
    def has_content(self):
        if self.logs:
            return True
        else:
            return False

    @property
    def nav_subtitle(self):
        if self.logs:
            return "%d" % (len(self.logs))

    def render_content(self, request):
        if not self.logs:
            return 'No logs in request.'

        stats = {}
        for api_call in ('get', 'get_multi', 'set', 'set_multi', 'delete', 'delete_multi', ):
            stats[api_call] = {
                'hit': 0,
                'miss': 0,
                'fractional-hit': 0,
                'fractional-miss': 0,
                'total': 0,
                'total_time': 0,
            }

        for r in self.logs:
            stats[r[0]]['total'] += 1
            stats[r[0]]['total_time'] += r[1]
            # dk - dogpile key
            # dr - dogpile result
            if r[0] == 'get_multi':
                for (dk, dr) in r[2]:
                    if dr is True:
                        stats[r[0]]['fractional-hit'] += 1
                    elif dr is False:
                        stats[r[0]]['fractional-miss'] += 1
            else:
                for (dk, dr) in r[2]:
                    if dr is True:
                        stats[r[0]]['hit'] += 1
                    elif dr is False:
                        stats[r[0]]['miss'] += 1

        self.data = {
            'logs': self.logs,
            'stats': stats,
        }
        return super(DogpileDebugPanel, self).render_content(request)

    def render_vars(self, request):
        return {
            'static_path': request.static_url(STATIC_PATH),
            'root_path': request.route_url(ROOT_ROUTE_NAME)
        }



# stdlib
import logging
import time

# pyramid
from pyramid.threadlocal import get_current_request

# dogpile
import dogpile
from dogpile.cache.api import NO_VALUE, CachedValue
from dogpile.cache.proxy import ProxyBackend


# local
from .panels.pyramid_dogpile import DogpileDebugPanel


def includeme(config):
    config.registry.settings['debugtoolbar.panels'].append(DogpileDebugPanel)
    if 'mako.directories' not in config.registry.settings:
        config.registry.settings['mako.directories'] = []

    # custom property || app_domain / DomainObject
    config.add_request_method(
        'pyramid_debugtoolbar_dogpile.setup_dogpile_logging',
        'dogpile_logging',
        reify=True,
    )


def setup_dogpile_logging(request):
    """
        calls_raw =  ordered list of calls
    """
    logging.debug('initializing setup_dogpile_logging')
    return {
        'api_calls': [],  # (api_call, duration, (k + result))
    }


class LoggingProxy(ProxyBackend):
    """
    This is an instance of ProxyBackend
    It times the performance of the backend and logs if we see cache hits or misses
    """

    def set(self, key, value):
        _s = time.time()
        self.proxied.set(key, value)
        _f = time.time()
        _d = _f - _s
        r = get_current_request()
        if r:
            r.dogpile_logging['api_calls'].append(("set", _d, [(key, None), ]))

    def set_multi(self, mapping):
        _s = time.time()
        self.proxied.set_multi(mapping)
        _f = time.time()
        _d = _f - _s

        r = get_current_request()
        if r:
            _kvs = [(k, None) for k in mapping.keys()]
            r.dogpile_logging['api_calls'].append(("set_multi", _d, _kvs))

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def get(self, key):
        _s = time.time()
        v = self.proxied.get(key)
        _f = time.time()
        _d = _f - _s

        r = get_current_request()
        if r:
            r.dogpile_logging['api_calls'].append(("get", _d, [(key, True if v is not NO_VALUE else False), ]))
        return v

    def get_multi(self, keys):
        _s = time.time()
        vs = self.proxied.get_multi(keys)
        _f = time.time()
        _d = _f - _s

        r = get_current_request()
        if r:
            _results = [(True if v is not NO_VALUE else False) for v in vs]
            _dictionary = dict(zip(keys, _results))
            r.dogpile_logging['api_calls'].append(("get_multi", _d, _dictionary.items()))
        return vs

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def delete(self, key):
        _s = time.time()
        self.proxied.delete(key)
        _f = time.time()
        _d = _f - _s
        r = get_current_request()
        if r:
            r.dogpile_logging['api_calls'].append(("delete", _d, [(key, None), ]))

    def delete_multi(self, keys):
        _s = time.time()
        self.proxied.delete_multi(keys)
        _f = time.time()
        _d = _f - _s

        r = get_current_request()
        if r:
            _kvs = [(k, None) for k in keys]
            r.dogpile_logging['api_calls'].append(("delete_multi", _d, _kvs))



"""pyramid_debugtoolbar_dogpile installation script.
"""
import os
from setuptools import setup
from setuptools import find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, "README.md")).read()
README = README.split("\n\n", 1)[0] + "\n"

requires = ['pyramid_debugtoolbar>=2.2', ]

setup(
    name="pyramid_debugtoolbar_dogpile",
    author="Jonathan Vanasco",
    author_email="jonathan@findmeon.com",
    url="https://github.com/jvanasco/pyramid_debugtoolbar_dogpile",
    version="0.0.5",
    description="dogpile support for pyramid_debugtoolbar",
    keywords="web pyramid",
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        "Framework :: Pyramid",
        "Programming Language :: Python",
        "License :: OSI Approved :: MIT License",
    ],
    long_description=README,
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=requires,
    test_suite="tests",
)
